package org.vendor.tftp.tool;

package .var.folders.cl.9twcgjfx4vsg5zl73c6wxmvm0000gn.T;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import org.apache.commons.net.io.FromNetASCIIOutputStream;
import org.apache.commons.net.io.ToNetASCIIInputStream;
import org.apache.commons.net.tftp.TFTP;
import org.apache.commons.net.tftp.TFTPAckPacket;
import org.apache.commons.net.tftp.TFTPDataPacket;
import org.apache.commons.net.tftp.TFTPErrorPacket;
import org.apache.commons.net.tftp.TFTPPacket;
import org.apache.commons.net.tftp.TFTPPacketException;
import org.apache.commons.net.tftp.TFTPReadRequestPacket;
import org.apache.commons.net.tftp.TFTPWriteRequestPacket;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.vendor.tftp.common.DataPacket;
import org.vendor.tftp.common.OptionAck;
import org.vendor.tftp.common.TFTPCommon;
import org.vendor.tftp.common.TFTPOption;
import org.vendor.tftp.prefences.TFTPPreferences;
import org.vendor.tftp.tool.TFTPService;
import org.vendor.tftp.tool.TftpInfo;

public class TFTPTransfer implements Runnable {
  TFTPCommon transferTftp = null;
  
  private Logger logger = LogManager.getLogger(org.vendor.tftp.tool.TFTPTransfer.class);
  
  private int tftpTimeOut = TFTPPreferences.getTimeout();
  
  private int TFTP_MAXRETRIES = 50;
  
  private TFTPPacket tftpPacket;
  
  private int socketTimeout;
  
  private int maxTimeoutRetries;
  
  private File serverDirectory;
  
  private final HashSet<org.vendor.tftp.tool.TFTPTransfer> transfers;
  
  private int receiveNumber = 0;
  
  public TFTPTransfer(TFTPPacket tftpPacket, int socketTimeout, int maxTimeoutRetries, File serverDirectory, HashSet<org.vendor.tftp.tool.TFTPTransfer> transfers) {
    this.socketTimeout = socketTimeout;
    this.tftpPacket = tftpPacket;
    this.serverDirectory = serverDirectory;
    this.transfers = transfers;
  }
  
  public void shutdown() {
    try {
      this.transferTftp.close();
    } catch (RuntimeException e) {
      e.printStackTrace();
    } 
  }
  
  public void run() {
    try {
      this.logger.info("TFTPTransfer run...");
      this.transferTftp = new TFTPCommon();
      this.transferTftp.beginBufferedOps();
      this.transferTftp.setDefaultTimeout(this.socketTimeout);
      this.transferTftp.open();
      this.logger.info("TFTPTransfer open");
      this.logger.info("tftpTimeout: {}", Integer.valueOf(this.tftpTimeOut));
      this.logger.info("winsize block count:" + TFTPService.getWinSizeBlockCount());
      if (this.tftpPacket instanceof TFTPReadRequestPacket) {
        this.logger.info("TFTPTransfer handle Read");
        handleRead((TFTPReadRequestPacket)this.tftpPacket);
      } else if (this.tftpPacket instanceof TFTPWriteRequestPacket) {
        this.logger.info("TFTPTransfer handle Write");
        handleWrite((TFTPWriteRequestPacket)this.tftpPacket);
      } 
    } catch (IOException e) {
      e.printStackTrace();
    } catch (TFTPPacketException e) {
      e.printStackTrace();
    } finally {
      if (this.transferTftp != null && this.transferTftp.isOpen()) {
        this.logger.info("transferTftp close.");
        this.transferTftp.endBufferedOps();
        this.transferTftp.close();
      } 
      synchronized (this.transfers) {
        this.transfers.remove(this);
      } 
    } 
    this.logger.info("TFTPTransfer end.");
  }
  
  private void handleRead(TFTPReadRequestPacket trrp) {
    ToNetASCIIInputStream toNetASCIIInputStream;
    int blksize = TFTPService.getTftpBlockSize();
    int timeout = TFTPService.getTftpTimeout();
    this.logger.info("blksize:" + blksize);
    this.logger.info("timeout:" + timeout);
    InputStream bis = null;
    try {
      bis = new BufferedInputStream(new FileInputStream(buildFile(this.serverDirectory, trrp
              .getFilename(), false)));
      int fileLen = bis.available();
      if (trrp.getMode() == 0)
        toNetASCIIInputStream = new ToNetASCIIInputStream(bis); 
      try {
        this.logger.info("send Option Oack.");
        sendOptionOack(toNetASCIIInputStream.available(), trrp.getAddress(), trrp
            .getPort(), blksize, timeout, 0);
        sendFileToClientByWindowSize(trrp, blksize, (InputStream)toNetASCIIInputStream, fileLen);
        this.logger.info("send data end.");
      } catch (TFTPPacketException tftpPacketException) {
        this.logger.error("timeout:" + tftpPacketException.getMessage());
      } 
    } catch (FileNotFoundException e) {
      try {
        this.transferTftp.bufferedSend((TFTPPacket)new TFTPErrorPacket(trrp
              .getAddress(), trrp.getPort(), 1, e
              .getMessage()));
      } catch (IOException ioException) {
        this.logger.error("timeout:" + ioException.getMessage());
      } 
      return;
    } catch (IOException ioException) {
      this.logger.error("timeout:" + ioException.getMessage());
    } finally {
      try {
        if (toNetASCIIInputStream != null)
          toNetASCIIInputStream.close(); 
      } catch (IOException e) {
        this.logger.error("timeout:" + e.getMessage());
      } 
      this.logger.info("handle Read file :" + trrp.getFilename() + " close.");
    } 
  }
  
  private void checkTFTPAckPacket(List<DatagramPacket> winSizeBufferDataList, int expectBlockNumber) throws IOException {
    this.receiveNumber = 0;
    int retransmits = 0;
    TFTPPacket packet = null;
    while (true) {
      try {
        packet = this.transferTftp.bufferedReceive();
      } catch (InterruptedIOException interruptedIOException) {
      
      } catch (SocketException e) {
        this.logger.error(e.getMessage());
      } catch (IOException e) {
        this.logger.error(e.getMessage());
      } catch (TFTPPacketException e) {
        this.logger.error(e.getMessage());
      } 
      if (packet != null && 
        checkAckPacket(winSizeBufferDataList, expectBlockNumber, packet))
        break; 
      if (packet == null || this.receiveNumber > 1) {
        this.logger.info("checkTFTPAckPacket");
        this.logger.info("expectBlockNumber" + expectBlockNumber);
        resendPacket(winSizeBufferDataList);
        if (retransmits++ > this.TFTP_MAXRETRIES)
          throw new IOException(getClient(this.transferTftp) + " Maximum retransmit count exceeded"); 
        this.logger.info("retransmits:" + retransmits);
      } 
    } 
  }
  
  private boolean checkAckPacket(List<DatagramPacket> winSizeBufferDataList, int expectBlockNumber, TFTPPacket packet) {
    if (packet instanceof TFTPAckPacket) {
      TFTPAckPacket ack = (TFTPAckPacket)packet;
      if (ack.getBlockNumber() == 0 && expectBlockNumber == 0) {
        if (winSizeBufferDataList != null)
          TftpInfo.curAckIndex++; 
        return true;
      } 
      if (expectBlockNumber == ack.getBlockNumber()) {
        TftpInfo.curAckIndex++;
        int length = ((DatagramPacket)winSizeBufferDataList.get(0)).getLength() - 4;
        TftpInfo.totalBytes += length;
        winSizeBufferDataList.remove(0);
        return true;
      } 
      this.receiveNumber++;
    } else {
      if (packet instanceof TFTPDataPacket)
        return true; 
      return false;
    } 
    return false;
  }
  
  private void resendPacket(List<DatagramPacket> winSizeBufferDataList) throws IOException {
    if (winSizeBufferDataList != null && winSizeBufferDataList.size() > 0) {
      for (DatagramPacket sentData : winSizeBufferDataList) {
        this.transferTftp.sendDataPacket(sentData);
        this.logger.info("send data blockNumber: {}", Integer.valueOf(DataPacket.getBlockNumber(sentData)));
      } 
      this.receiveNumber = 0;
    } 
  }
  
  public static String getClient(TFTP tftpSock) {
    InetAddress inetAddress = tftpSock.getLocalAddress();
    int localPort = tftpSock.getLocalPort();
    String clientStr = "";
    if (inetAddress != null)
      clientStr = clientStr + clientStr; 
    if (localPort != 0) {
      clientStr = clientStr + ":";
      clientStr = clientStr + clientStr;
    } 
    return clientStr;
  }
  
  private void sendFileToClientByWindowSize(TFTPReadRequestPacket trrp, int blksize, InputStream bis, int fileLen) throws SocketException, IOException, InterruptedIOException, TFTPPacketException {
    TftpInfo.curAckIndex = 1;
    TftpInfo.packetSize = blksize;
    if (TFTPService.isUsb2EtherNet()) {
      TftpInfo.winSize = 1;
    } else {
      TftpInfo.winSize = TFTPService.getWinSizeBlockCount();
    } 
    TftpInfo.totalBytes = 0;
    TftpInfo.lastToSend = 1;
    TftpInfo.timeOut = 0;
    TftpInfo.bytesByCountMap.clear();
    List<DatagramPacket> winSizeBufferDataList = new ArrayList<>();
    TftpInfo.lastBlockOfFile = fileLen / TftpInfo.packetSize;
    TftpInfo.lastBlockOfFile++;
    if (fileLen / TftpInfo.packetSize > 0)
      TftpInfo.lastBlockOfFile++; 
    this.logger.error("lastBlockOfFile" + TftpInfo.lastBlockOfFile);
    byte[] data = null;
    int read = 0;
    do {
      updateSocketTimeout();
      DataPacket dataPacket = new DataPacket();
      for (; TftpInfo.lastToSend <= Math.min(TftpInfo.curAckIndex + TftpInfo.winSize - 1, TftpInfo.lastBlockOfFile); 
        TftpInfo.lastToSend++) {
        if (TftpInfo.totalBytes + TftpInfo.packetSize > fileLen) {
          int remainLen = fileLen - TftpInfo.totalBytes;
          data = new byte[remainLen];
        } else {
          data = new byte[TftpInfo.packetSize];
        } 
        read = bis.read(data);
        TftpInfo.bytesByCountMap.put(Integer.valueOf(TftpInfo.lastToSend), Integer.valueOf(read));
        if (read < 0)
          read = 0; 
        int blockNumber = TftpInfo.lastToSend % 65536;
        DatagramPacket sentData = dataPacket.newDatagram(trrp.getAddress(), trrp.getPort(), blockNumber, data, read);
        this.transferTftp.sendDataPacket(sentData);
        winSizeBufferDataList.add(sentData);
      } 
      int expectBlockNumber = TftpInfo.curAckIndex % 65536;
      if (TftpInfo.lastToSend > TftpInfo.lastBlockOfFile)
        break; 
      checkTFTPAckPacket(winSizeBufferDataList, expectBlockNumber);
    } while (TftpInfo.lastToSend <= TftpInfo.lastBlockOfFile);
  }
  
  public int getTimeout() {
    return this.tftpTimeOut;
  }
  
  private void updateSocketTimeout() throws SocketException {
    int socketTimeOut = 0;
    switch (TftpInfo.timeOut) {
      case 0:
        socketTimeOut = (getTimeout() + 3) / 4 * 1000;
        break;
      case 1:
        socketTimeOut = (getTimeout() + 1) / 2 * 1000;
        break;
      default:
        socketTimeOut = getTimeout() * 1000;
        break;
    } 
    this.transferTftp.setDefaultTimeout(socketTimeOut);
    this.transferTftp.setSoTimeout(socketTimeOut);
  }
  
  private void sendOptionOack(int tSize, InetAddress address, int port, int blksize, int timeout, int expectBlockNumber) throws IOException, TFTPPacketException {
    TFTPOption options = new TFTPOption();
    options.setTimeOut(timeout);
    options.setBlkSize(blksize);
    options.settSize(tSize);
    this.logger.info("option acknowledgement.");
    this.logger.info("timeout:" + timeout);
    this.logger.info("blksize:" + blksize);
    this.logger.info("tSize:" + tSize);
    OptionAck oack = new OptionAck(options);
    byte[] AcknowledgementData = oack.getAcknowledgementData();
    DatagramPacket dategramPacket = new DatagramPacket(AcknowledgementData, AcknowledgementData.length, address, port);
    TFTPPacket tftpPacket = TFTPDataPacket.newTFTPPacket(dategramPacket);
    dategramPacket.setData(oack.getBytes());
    this.logger.info("Creating dategramPacket Packets.");
    this.logger.info("AcknowledgementData:" + AcknowledgementData.length);
    this.logger.info("InetAddress:" + address.getHostAddress());
    this.logger.info("expectBlockNumber:" + expectBlockNumber);
    this.logger.info("send OACK option information.");
    this.transferTftp.sendTftpPacket(tftpPacket, dategramPacket);
    if (expectBlockNumber == 0)
      checkTFTPAckPacket(this.transferTftp, tftpPacket, dategramPacket, expectBlockNumber); 
  }
  
  private void checkTFTPAckPacket(TFTPCommon tftp, TFTPPacket tftpPacket, DatagramPacket dategramPacket, int expectBlockNumber) throws IOException {
    this.receiveNumber = 0;
    int retransmits = 0;
    TFTPPacket tftpAckPacket = null;
    while (true) {
      try {
        tftpAckPacket = this.transferTftp.bufferedReceive();
      } catch (InterruptedIOException e) {
        this.logger.error(e.getMessage());
      } catch (SocketException e) {
        this.logger.error(e.getMessage());
      } catch (IOException e) {
        this.logger.error(e.getMessage());
      } catch (TFTPPacketException e) {
        this.logger.error(e.getMessage());
      } 
      if (tftpAckPacket != null)
        if (tftpAckPacket instanceof TFTPAckPacket) {
          TFTPAckPacket ack = (TFTPAckPacket)tftpAckPacket;
          if (expectBlockNumber == ack.getBlockNumber())
            break; 
          this.receiveNumber++;
        } else if (tftpAckPacket instanceof TFTPDataPacket) {
          break;
        }  
      if (tftpAckPacket == null || this.receiveNumber > 1) {
        this.logger.info("checkTFTPAckPacket");
        tftp.sendTftpPacket(tftpPacket, dategramPacket);
        this.receiveNumber = 0;
        if (retransmits++ > this.TFTP_MAXRETRIES)
          throw new IOException(getClient(this.transferTftp) + " Maximum retransmit count exceeded"); 
        this.logger.info("retransmits: " + retransmits);
      } 
    } 
  }
  
  private void handleWrite(TFTPWriteRequestPacket twrp) throws IOException, TFTPPacketException {
    FromNetASCIIOutputStream fromNetASCIIOutputStream;
    int blksize = TFTPService.getTftpBlockSize();
    int timeout = TFTPService.getTftpTimeout();
    sendOptionOack(0, twrp.getAddress(), twrp.getPort(), blksize, timeout, 1);
    OutputStream bos = null;
    try {
      int lastBlock = 0;
      String fileName = twrp.getFilename();
      try {
        File temp = buildFile(this.serverDirectory, fileName, true);
        bos = new BufferedOutputStream(new FileOutputStream(temp));
        if (twrp.getMode() == 0)
          fromNetASCIIOutputStream = new FromNetASCIIOutputStream(bos); 
      } catch (IOException e) {
        this.transferTftp.bufferedSend((TFTPPacket)new TFTPErrorPacket(twrp
              .getAddress(), twrp.getPort(), 0, e
              .getMessage()));
        return;
      } 
      while (true) {
        DataPacket dataTFTPPacket = null;
        int count = 0;
        while (dataTFTPPacket == null || 
          !dataTFTPPacket.getAddress().equals(twrp.getAddress()) || dataTFTPPacket
          .getPort() != twrp.getPort()) {
          if (dataTFTPPacket != null)
            this.transferTftp.bufferedSend((TFTPPacket)new TFTPErrorPacket(dataTFTPPacket
                  .getAddress(), dataTFTPPacket.getPort(), 5, "Data cannot be received")); 
          try {
            dataTFTPPacket = this.transferTftp.receiveDataPacket();
          } catch (SocketTimeoutException e) {
            if (count >= this.maxTimeoutRetries)
              throw e; 
            TFTPAckPacket lastSentAck = new TFTPAckPacket(twrp.getAddress(), twrp.getPort(), 1);
            this.transferTftp.bufferedSend((TFTPPacket)lastSentAck);
            count++;
          } 
        } 
        if (dataTFTPPacket != null && dataTFTPPacket
          .getPacketType() == TFTPCommon.PacketType.WRITE_REQUEST.ordinal()) {
          TFTPAckPacket lastSentAck = new TFTPAckPacket(twrp.getAddress(), twrp.getPort(), 0);
          this.transferTftp.bufferedSend((TFTPPacket)lastSentAck);
          continue;
        } 
        if (dataTFTPPacket == null || dataTFTPPacket
          .getPacketType() != TFTPCommon.PacketType.DATA.ordinal())
          break; 
        int block = dataTFTPPacket.getBlockNumber();
        byte[] data = dataTFTPPacket.getData();
        int dataLength = dataTFTPPacket.getDataLength();
        int dataOffset = dataTFTPPacket.getDataOffset();
        if (block > lastBlock || (lastBlock == 65535 && block == 0)) {
          fromNetASCIIOutputStream.write(data, dataOffset, dataLength);
          lastBlock = block;
        } 
        TFTPAckPacket last = new TFTPAckPacket(twrp.getAddress(), twrp.getPort(), block);
        this.transferTftp.bufferedSend((TFTPPacket)last);
        if (dataLength < blksize) {
          fromNetASCIIOutputStream.close();
          for (int i = 0; i < this.maxTimeoutRetries; i++) {
            try {
              dataTFTPPacket = this.transferTftp.receiveDataPacket();
            } catch (SocketTimeoutException e) {
              break;
            } 
            if (dataTFTPPacket != null && (
              !dataTFTPPacket.getAddress().equals(twrp
                .getAddress()) || dataTFTPPacket
              .getPort() != twrp.getPort())) {
              this.transferTftp.bufferedSend((TFTPPacket)new TFTPErrorPacket(dataTFTPPacket
                    .getAddress(), dataTFTPPacket
                    .getPort(), 5, "Unexpected Host or Port"));
            } else {
              this.transferTftp.bufferedSend((TFTPPacket)last);
            } 
          } 
          break;
        } 
      } 
    } finally {
      if (fromNetASCIIOutputStream != null)
        fromNetASCIIOutputStream.close(); 
    } 
  }
  
  private File buildFile(File serverDirectory, String fileName, boolean isDirectory) throws IOException {
    File temp = new File(serverDirectory, fileName);
    File calFile = temp.getCanonicalFile();
    if (isDirectory)
      createDirectory(calFile.getParentFile()); 
    return calFile;
  }
  
  private void createDirectory(File file) throws IOException {
    File parentFile = file.getParentFile();
    if (parentFile != null) {
      if (!parentFile.exists())
        createDirectory(parentFile); 
      if (parentFile.isDirectory()) {
        if (file.isDirectory())
          return; 
        if (file.mkdir())
          throw new IOException("Unable to create file, insufficient disk space, or no write permission."); 
      } 
    } 
  }
}
